@@ -1,5 +1,5 @@
 # mssql/pyodbc.py
-# Copyright (C) 2005-2018 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2019 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -55,11 +55,11 @@ Pass through exact Pyodbc string
 
 A PyODBC connection string can also be sent exactly as specified in
 `ConnectionStrings <https://code.google.com/p/pyodbc/wiki/ConnectionStrings>`_
-into the driver using the parameter ``odbc_connect``.  The delimeters must be URL escaped, however,
-as illustrated below using ``urllib.quote_plus``::
+into the driver using the parameter ``odbc_connect``.  The delimeters must be
+URL escaped, however, as illustrated below using ``urllib.parse.quote_plus``::
 
     import urllib
-    params = urllib.quote_plus("DRIVER={SQL Server Native Client 10.0};SERVER=dagger;DATABASE=test;UID=user;PWD=password")
+    params = urllib.parse.quote_plus("DRIVER={SQL Server Native Client 10.0};SERVER=dagger;DATABASE=test;UID=user;PWD=password")
 
     engine = create_engine("mssql+pyodbc:///?odbc_connect=%s" % params)
 
@@ -85,14 +85,46 @@ Pyodbc only has partial support for rowcount.  See the notes at
 :ref:`mssql_rowcount_versioning` for important notes when using ORM
 versioning.
 
-"""
+.. _mssql_pyodbc_fastexecutemany:
+
+Fast Executemany Mode
+---------------------
+
+The Pyodbc driver has added support for a "fast executemany" mode of execution
+which greatly reduces round trips for a DBAPI ``executemany()`` call when using
+Microsoft ODBC drivers.  The feature is enabled by setting the flag
+``.fast_executemany`` on the DBAPI cursor when an executemany call is to be
+used.   The SQLAlchemy pyodbc SQL Server dialect supports setting this flag
+automatically when the ``.fast_executemany`` flag is passed to
+:func:`.create_engine`; note that the ODBC driver must be the Microsoft driver
+in order to use this flag::
+
+    engine = create_engine(
+        "mssql+pyodbc://scott:tiger@mssql2017:1433/test?driver=ODBC+Driver+13+for+SQL+Server",
+        fast_executemany=True)
+
+.. versionadded:: 1.3
+
+.. seealso::
+
+    `fast executemany <https://github.com/mkleehammer/pyodbc/wiki/Features-beyond-the-DB-API#fast_executemany>`_
+    - on github
+
+
+"""  # noqa
 
-from .base import MSExecutionContext, MSDialect, BINARY, VARBINARY
-from ...connectors.pyodbc import PyODBCConnector
-from ... import types as sqltypes, util, exc
 import decimal
 import re
 
+from .base import BINARY
+from .base import MSDialect
+from .base import MSExecutionContext
+from .base import VARBINARY
+from ... import exc
+from ... import types as sqltypes
+from ... import util
+from ...connectors.pyodbc import PyODBCConnector
+
 
 class _ms_numeric_pyodbc(object):
 
@@ -105,16 +137,13 @@ class _ms_numeric_pyodbc(object):
 
     def bind_processor(self, dialect):
 
-        super_process = super(_ms_numeric_pyodbc, self).\
-            bind_processor(dialect)
+        super_process = super(_ms_numeric_pyodbc, self).bind_processor(dialect)
 
         if not dialect._need_decimal_fix:
             return super_process
 
         def process(value):
-            if self.asdecimal and \
-                    isinstance(value, decimal.Decimal):
-
+            if self.asdecimal and isinstance(value, decimal.Decimal):
                 adjusted = value.adjusted()
                 if adjusted < 0:
                     return self._small_dec_to_string(value)
@@ -125,6 +154,7 @@ class _ms_numeric_pyodbc(object):
                 return super_process(value)
             else:
                 return value
+
         return process
 
     # these routines needed for older versions of pyodbc.
@@ -132,30 +162,31 @@ class _ms_numeric_pyodbc(object):
 
     def _small_dec_to_string(self, value):
         return "%s0.%s%s" % (
-            (value < 0 and '-' or ''),
-            '0' * (abs(value.adjusted()) - 1),
-            "".join([str(nint) for nint in value.as_tuple()[1]]))
+            (value < 0 and "-" or ""),
+            "0" * (abs(value.adjusted()) - 1),
+            "".join([str(nint) for nint in value.as_tuple()[1]]),
+        )
 
     def _large_dec_to_string(self, value):
         _int = value.as_tuple()[1]
-        if 'E' in str(value):
+        if "E" in str(value):
             result = "%s%s%s" % (
-                (value < 0 and '-' or ''),
+                (value < 0 and "-" or ""),
                 "".join([str(s) for s in _int]),
-                "0" * (value.adjusted() - (len(_int) - 1)))
+                "0" * (value.adjusted() - (len(_int) - 1)),
+            )
         else:
             if (len(_int) - 1) > value.adjusted():
                 result = "%s%s.%s" % (
-                    (value < 0 and '-' or ''),
-                    "".join(
-                        [str(s) for s in _int][0:value.adjusted() + 1]),
-                    "".join(
-                        [str(s) for s in _int][value.adjusted() + 1:]))
+                    (value < 0 and "-" or ""),
+                    "".join([str(s) for s in _int][0 : value.adjusted() + 1]),
+                    "".join([str(s) for s in _int][value.adjusted() + 1 :]),
+                )
             else:
                 result = "%s%s" % (
-                    (value < 0 and '-' or ''),
-                    "".join(
-                        [str(s) for s in _int][0:value.adjusted() + 1]))
+                    (value < 0 and "-" or ""),
+                    "".join([str(s) for s in _int][0 : value.adjusted() + 1]),
+                )
         return result
 
 
@@ -186,6 +217,7 @@ class _ms_binary_pyodbc(object):
             else:
                 # pyodbc-specific
                 return dialect.dbapi.BinaryNull
+
         return process
 
 
@@ -217,9 +249,11 @@ class MSExecutionContext_pyodbc(MSExecutionContext):
 
         # don't embed the scope_identity select into an
         # "INSERT .. DEFAULT VALUES"
-        if self._select_lastrowid and \
-                self.dialect.use_scope_identity and \
-                len(self.parameters[0]):
+        if (
+            self._select_lastrowid
+            and self.dialect.use_scope_identity
+            and len(self.parameters[0])
+        ):
             self._embedded_scope_identity = True
 
             self.statement += "; select scope_identity()"
@@ -235,7 +269,7 @@ class MSExecutionContext_pyodbc(MSExecutionContext):
                     # without closing it (FreeTDS particularly)
                     row = self.cursor.fetchall()[0]
                     break
-                except self.dialect.dbapi.Error as e:
+                except self.dialect.dbapi.Error:
                     # no way around this - nextset() consumes the previous set
                     # so we need to just keep flipping
                     self.cursor.nextset()
@@ -255,54 +289,82 @@ class MSDialect_pyodbc(PyODBCConnector, MSDialect):
             sqltypes.Numeric: _MSNumeric_pyodbc,
             sqltypes.Float: _MSFloat_pyodbc,
             BINARY: _BINARY_pyodbc,
-
             # SQL Server dialect has a VARBINARY that is just to support
             # "deprecate_large_types" w/ VARBINARY(max), but also we must
             # handle the usual SQL standard VARBINARY
             VARBINARY: _VARBINARY_pyodbc,
             sqltypes.VARBINARY: _VARBINARY_pyodbc,
             sqltypes.LargeBinary: _VARBINARY_pyodbc,
-        }
+        },
     )
 
-    def __init__(self, description_encoding=None, **params):
-        if 'description_encoding' in params:
-            self.description_encoding = params.pop('description_encoding')
+    def __init__(
+        self, description_encoding=None, fast_executemany=False, **params
+    ):
+        if "description_encoding" in params:
+            self.description_encoding = params.pop("description_encoding")
         super(MSDialect_pyodbc, self).__init__(**params)
-        self.use_scope_identity = self.use_scope_identity and \
-            self.dbapi and \
-            hasattr(self.dbapi.Cursor, 'nextset')
-        self._need_decimal_fix = self.dbapi and \
-            self._dbapi_version() < (2, 1, 8)
+        self.use_scope_identity = (
+            self.use_scope_identity
+            and self.dbapi
+            and hasattr(self.dbapi.Cursor, "nextset")
+        )
+        self._need_decimal_fix = self.dbapi and self._dbapi_version() < (
+            2,
+            1,
+            8,
+        )
+        self.fast_executemany = fast_executemany
 
     def _get_server_version_info(self, connection):
         try:
+            # "Version of the instance of SQL Server, in the form
+            # of 'major.minor.build.revision'"
             raw = connection.scalar(
-                "SELECT CAST(SERVERPROPERTY('ProductVersion') AS VARCHAR)")
+                "SELECT CAST(SERVERPROPERTY('ProductVersion') AS VARCHAR)"
+            )
         except exc.DBAPIError:
             # SQL Server docs indicate this function isn't present prior to
             # 2008.  Before we had the VARCHAR cast above, pyodbc would also
             # fail on this query.
-            return super(MSDialect_pyodbc, self).\
-                _get_server_version_info(connection)
+            return super(MSDialect_pyodbc, self)._get_server_version_info(
+                connection, allow_chars=False
+            )
         else:
             version = []
-            r = re.compile(r'[.\-]')
+            r = re.compile(r"[.\-]")
             for n in r.split(raw):
                 try:
                     version.append(int(n))
                 except ValueError:
-                    version.append(n)
+                    pass
             return tuple(version)
 
+    def do_executemany(self, cursor, statement, parameters, context=None):
+        if self.fast_executemany:
+            cursor.fast_executemany = True
+        super(MSDialect_pyodbc, self).do_executemany(
+            cursor, statement, parameters, context=context
+        )
+
     def is_disconnect(self, e, connection, cursor):
         if isinstance(e, self.dbapi.Error):
             for code in (
-                    '08S01', '01002', '08003', '08007',
-                    '08S02', '08001', 'HYT00', 'HY010'):
+                "08S01",
+                "01002",
+                "08003",
+                "08007",
+                "08S02",
+                "08001",
+                "HYT00",
+                "HY010",
+                "10054",
+            ):
                 if code in str(e):
                     return True
         return super(MSDialect_pyodbc, self).is_disconnect(
-            e, connection, cursor)
+            e, connection, cursor
+        )
+
 
 dialect = MSDialect_pyodbc
